/* Language Abstraction Editor ver.0.1.0
 * 
 * This software is in the public domain.
 * There are no restrictions on any sort of usage of this software.
 * 
 * $lae: laewin.cpp,v 1.19.9 2001/09/18 15:39:21 Toshihiro Inoue Exp $
 */

#include "laewin.h"

#include "laetree.h"

#include <ffclist.h>
#include <ffcstring.h>

#include "main.h"

FXDEFMAP(LAEWin) LAEWinMap[] = {
  FXMAPFUNC(SEL_CLOSE, 0, LAEWin::onClose),
  FXMAPFUNC(SEL_COMMAND, LAEWin::ID_CLOSE_DOCUMENT, LAEWin::onClose),
  FXMAPFUNC(SEL_COMMAND, LAEWin::ID_NEW , LAEWin::onMnuNew),
  FXMAPFUNC(SEL_COMMAND, LAEWin::ID_OPEN, LAEWin::onMnuOpen),
  FXMAPFUNC(SEL_COMMAND, LAEWin::ID_SAVE, LAEWin::onMnuSave),
  FXMAPFUNC(SEL_COMMAND, LAEWin::ID_SAVEAS, LAEWin::onMnuSaveAs),
  FXMAPFUNC(SEL_COMMAND, LAEWin::ID_ABOUT, LAEWin::onMnuAbout),
  FXMAPFUNC(SEL_COMMAND, LAEWin::ID_TAB, LAEWin::onSetCurrent),
};

FXIMPLEMENT(LAEWin, FXTopWindow, LAEWinMap, ARRAYNUMBER(LAEWinMap))

static int windowCount = 0;

#ifndef WIN32
const FXchar* patterns[] = {
  "LAE Files", "*.lae",
  "All Files", "*",
  NULL
};
#else
#include <windows.h>
const FXchar* patterns =
  "LAE Files (*.lae)\0" "*.lae\0"
  "All Files (*.*)\0" "*.*\0";
#endif

static const char* fileNew[] = {
"16 16 3 1",
"         c None",
"+        c #000000",
"@        c #FFFFFF",
"                ",
"   ++++++++     ",
"   +@@@@@@++    ",
"   +@@@@@@+@+   ",
"   +@@@@@@++++  ",
"   +@@@@@@@@@+  ",
"   +@@@@@@@@@+  ",
"   +@@@@@@@@@+  ",
"   +@@@@@@@@@+  ",
"   +@@@@@@@@@+  ",
"   +@@@@@@@@@+  ",
"   +@@@@@@@@@+  ",
"   +@@@@@@@@@+  ",
"   +++++++++++  ",
"                ",
"                "};

static const char* fileOpen[] = {
"16 16 5 1",
"         c None",
"+        c #000000",
"@        c #FFFF00",
"#        c #FFFFFF",
"$        c #848400",
"                ",
"         +++    ",
"        +   + + ",
"             ++ ",
" +++        +++ ",
"+@#@+++++++     ",
"+#@#@#@#@#+     ",
"+@#@#@#@#@+     ",
"+#@#@+++++++++++",
"+@#@+$$$$$$$$$+ ",
"+#@+$$$$$$$$$+  ",
"+@+$$$$$$$$$+   ",
"++$$$$$$$$$+    ",
"+++++++++++     ",
"                ",
"                "};

static const char * fileSave[] = {
"16 16 4 1",
"         c None",
"+        c #000000",
"@        c #848400",
"#        c #C6C6C6",
"                ",
" ++++++++++++++ ",
" +@+########+#+ ",
" +@+########+++ ",
" +@+########+@+ ",
" +@+########+@+ ",
" +@+########+@+ ",
" +@+########+@+ ",
" +@@++++++++@@+ ",
" +@@@@@@@@@@@@+ ",
" +@@+++++++++@+ ",
" +@@++++++##+@+ ",
" +@@++++++##+@+ ",
" +@@++++++##+@+ ",
"  +++++++++++++ ",
"                "};

static FXXPMIcon* icoNew  = NULL;
static FXXPMIcon* icoOpen = NULL;
static FXXPMIcon* icoSave = NULL;


void LAEWin::__init_class__()
{
  tshMenu = 0;
  menu = 0;
  mnuFile = 0;
  mnuEdit = 0;
  mnuItem = 0;
  mnuHelp = 0;
  tshTool = 0;
  tool = 0;
  tab1 = 0;
  tab1i1 = 0;
  tree = 0;
  tab1i2 = 0;
  codeC = 0;
  tab1i3 = 0;
  codePerl = 0;
  tab1i4 = 0;
  codePython = 0;
  tab1i5 = 0;
  codeRuby = 0;
  tab1i6 = 0;
  codeVB = 0;
  tab1i7 = 0;
  codePascal = 0;
  tab1i8 = 0;
  codeTcl = 0;
  status = 0;
}


LAEWin::LAEWin(FXApp* app): FXTopWindow(app, APPNAME, NULL, NULL, DECOR_ALL, 30, 30, 400, 400, 0, 0, 0, 0, 0, 0)
{
  __init_class__();
  
  windowCount++;
  
  tshMenu = new FXToolbarShell(this, FRAME_RAISED);
  menu = new FXMenubar(this, tshMenu, FRAME_GROOVE | LAYOUT_SIDE_TOP | LAYOUT_FILL_X);
  new FXToolbarGrip(menu, menu, FXMenubar::ID_TOOLBARGRIP, TOOLBARGRIP_DOUBLE);
  
  tshTool = new FXToolbarShell(this, FRAME_RAISED);
  tool = new FXToolbar(this, tshTool, FRAME_GROOVE | LAYOUT_SIDE_TOP | LAYOUT_FILL_X);
  new FXToolbarGrip(tool, tool, FXToolbar::ID_TOOLBARGRIP, TOOLBARGRIP_DOUBLE);
  
  status = new FXStatusbar(this,LAYOUT_SIDE_BOTTOM|LAYOUT_FILL_X|STATUSBAR_WITH_DRAGCORNER);
  
  initIcon();
  initTab();
  initMenu();
  initTool();
  
  tree->setFocus();
}


void LAEWin::initIcon()
{
  if(!icoNew ) icoNew  = new FXXPMIcon(getApp(), fileNew );
  if(!icoOpen) icoOpen = new FXXPMIcon(getApp(), fileOpen);
  if(!icoSave) icoSave = new FXXPMIcon(getApp(), fileSave);
}


void LAEWin::initTab()
{
  tab1 = new FXTabBook(this, this,ID_TAB, LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,200,0, 0,0,0,0);
  
  tab1i1 = new FXTabItem(tab1, "Abstract", NULL);
  FXPacker* box = new FXPacker(tab1, FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  box = new FXPacker(box, FRAME_NORMAL|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0, 0,0);
  tree = new LAETree(
    box, 0, this,ID_TREE,
    LAYOUT_FILL_X | LAYOUT_FILL_Y
    | TREELIST_SHOWS_LINES | TREELIST_SHOWS_BOXES
    | TREELIST_ROOT_BOXES | TREELIST_EXTENDEDSELECT);
  
  tab1i2 = new FXTabItem(tab1, "C", NULL);
  box = new FXVerticalFrame(tab1, FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  box = new FXPacker(box, FRAME_NORMAL|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  codeC = new FXText(box, NULL,0, LAYOUT_FILL_X|LAYOUT_FILL_Y);
  
  tab1i3 = new FXTabItem(tab1, "Perl", NULL);
  box = new FXVerticalFrame(tab1, FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  box = new FXPacker(box, FRAME_NORMAL|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  codePerl = new FXText(box, NULL,0, LAYOUT_FILL_X|LAYOUT_FILL_Y);
  
  tab1i4 = new FXTabItem(tab1, "Python", NULL);
  box = new FXVerticalFrame(tab1, FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  box = new FXPacker(box, FRAME_NORMAL|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  codePython = new FXText(box, NULL,0, LAYOUT_FILL_X|LAYOUT_FILL_Y);
  
  tab1i5 = new FXTabItem(tab1, "Ruby", NULL);
  box = new FXVerticalFrame(tab1, FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  box = new FXPacker(box, FRAME_NORMAL|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  codeRuby = new FXText(box, NULL,0, LAYOUT_FILL_X|LAYOUT_FILL_Y);
  
  tab1i6 = new FXTabItem(tab1, "VB", NULL);
  box = new FXVerticalFrame(tab1, FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  box = new FXPacker(box, FRAME_NORMAL|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  codeVB = new FXText(box, NULL,0, LAYOUT_FILL_X|LAYOUT_FILL_Y);
  
  tab1i7 = new FXTabItem(tab1, "Pascal", NULL);
  box = new FXVerticalFrame(tab1, FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  box = new FXPacker(box, FRAME_NORMAL|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  codePascal = new FXText(box, NULL,0, LAYOUT_FILL_X|LAYOUT_FILL_Y);
  
  tab1i8 = new FXTabItem(tab1, "Tcl", NULL);
  box = new FXVerticalFrame(tab1, FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  box = new FXPacker(box, FRAME_NORMAL|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  codeTcl = new FXText(box, NULL,0, LAYOUT_FILL_X|LAYOUT_FILL_Y);
}


void LAEWin::initMenu()
{
  mnuFile = new FXMenuPane(this);
  new FXMenuTitle(menu, "&File", NULL, mnuFile);
  new FXMenuCommand(mnuFile, "&New\tCtl-N" , icoNew , this, ID_NEW);
  new FXMenuCommand(mnuFile, "&Open\tCtl-O", icoOpen, this, ID_OPEN);
  new FXMenuCommand(mnuFile, "&Save\tCtl-S", icoSave, this, ID_SAVE);
  new FXMenuCommand(mnuFile, "Save &As", NULL, this, ID_SAVEAS);
  new FXMenuCommand(mnuFile, "&Close\tCtl-W", NULL, this, ID_CLOSE_DOCUMENT);
  new FXMenuSeparator(mnuFile);
  new FXMenuCommand(mnuFile, "E&xit\tCtl-Q", NULL, getApp(), FXApp::ID_QUIT);
  
  mnuEdit = new FXMenuPane(this);
  new FXMenuTitle(menu, "&Edit", NULL, mnuEdit);
  new FXMenuCommand(mnuEdit, "Cu&t\tCtl-X"  , NULL, tree, FXTreeEditor::ID_CUT_SEL);
  new FXMenuCommand(mnuEdit, "&Copy\tCtl-C" , NULL, tree, FXTreeEditor::ID_COPY_SEL);
  new FXMenuCommand(mnuEdit, "&Paste\tCtl-V", NULL, tree, FXTreeEditor::ID_PASTE);
  new FXMenuCommand(mnuEdit, "&Delete", NULL, tree, FXTreeEditor::ID_DELETE);
  
  mnuItem = new FXMenuPane(this);
  new FXMenuTitle(menu, "&Item", NULL, mnuItem);
  new FXMenuCommand(mnuItem, "&Child" , NULL, tree, FXTreeEditor::ID_CHILD);
  new FXMenuCommand(mnuItem, "&Append", NULL, tree, FXTreeEditor::ID_APPEND);
  new FXMenuCommand(mnuItem, "&Insert", NULL, tree, FXTreeEditor::ID_INSERT);
  new FXMenuSeparator(mnuItem);
  new FXMenuCommand(mnuItem, "&Rename", NULL, tree, FXTreeEditor::ID_RENAME);
  new FXMenuCommand(mnuItem, "&Duplicate", NULL, tree, FXTreeEditor::ID_COPYITEM);
  
  mnuHelp = new FXMenuPane(this);
  new FXMenuTitle(menu, "&Help", NULL, mnuHelp);
  new FXMenuCommand(mnuHelp, "&About", NULL, this, ID_ABOUT);
}


void LAEWin::initTool()
{
  new FXButton(tool, "\tNew Ctl-N" , icoNew , this, ID_NEW , BUTTON_TOOLBAR|FRAME_RAISED);
  new FXButton(tool, "\tOpen Ctl-O", icoOpen, this, ID_OPEN, BUTTON_TOOLBAR|FRAME_RAISED);
  new FXButton(tool, "\tSave Ctl-S", icoSave, this, ID_SAVE, BUTTON_TOOLBAR|FRAME_RAISED);
}


void LAEWin::create()
{
  FXTopWindow::create();
  tshMenu->create();
  tshTool->create();
}


LAEWin::~LAEWin()
{
  delete mnuFile;
  delete mnuEdit;
  delete mnuItem;
  delete mnuHelp;
  delete menu;
  delete tshMenu;
  
  delete tool;
  delete tshTool;
  
  windowCount--;
  if(windowCount < 1) getApp()->exit();
}


/**
  Dummy for Message Map
*/

LAEWin::LAEWin()
{
  __init_class__();
  
}


long LAEWin::onClose(FXObject* sender, FXSelector sel, void* ptr)
{
  hide();
  delete this;
  return 1;
}


FXString LAEWin::execDlg(const FXString& title, FXbool save /*= false*/)
{
  FXString fn;
  FXbool flg = FALSE;
  
  fn = fileName;
  #ifndef WIN32
  FXFileDialog dlg(this, title);
  dlg.setPatternList(patterns);
  if(!fn.empty()) dlg.setFilename(fn);
  flg = dlg.execute();
  tree->setFocus();
  if(!flg) return "";
  return dlg.getFilename();
  #else
  OPENFILENAME ofn;
  memset(&ofn, 0, sizeof(OPENFILENAME));
  char buffer[MAX_PATH];
  #ifndef USE_UNICODE
  strcpy(buffer, fn.text());
  #else
  strcpy(buffer, utf2locale(fn).text());
  #endif
  ofn.lStructSize  = sizeof(OPENFILENAME);
  ofn.hwndOwner    = (HWND)xid;
  ofn.lpstrFilter  = patterns;
  ofn.nFilterIndex = 1;
  ofn.lpstrFile    = buffer;
  ofn.nMaxFile     = MAX_PATH;
  ofn.lpstrTitle   = title.text();
  ofn.lpstrDefExt  = "lae";
  ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
  if(save) {
    flg = GetSaveFileName(&ofn);
  } else {
    flg = GetOpenFileName(&ofn);
  }
  tree->setFocus();
  if(!flg) return "";
  #ifndef USE_UNICODE
  return ofn.lpstrFile;
  #else
  return locale2utf(ofn.lpstrFile);
  #endif
  #endif
}


long LAEWin::onMnuNew(FXObject* sender, FXSelector sel, void* ptr)
{
  LAEWin* w = new LAEWin(getApp());
  w->create();
  w->move(getX() + 20, getY() + 20);
  w->show();
  w->setFocus();
  return 1;
}


long LAEWin::onMnuOpen(FXObject* sender, FXSelector sel, void* ptr)
{
  FXString fn;
  
  fn = execDlg("Open File");
  if(fn.empty()) return 0;
  
  open(fn);
  return 1;
}


void LAEWin::open(const FXString& fn)
{
  FXFileStream stream;
  FXTreeItem* parent = 0;
  FXint indent = 0;
  
  if(!fileName.empty()) {
    LAEWin* w = new LAEWin(getApp());
    w->create();
    w->move(getX() + 20, getY() + 20);
    w->show();
    w->setFocus();
    w->open(fn);
  } else {
    tree->removeAllItems();
    fileName = fn;
    if(stream.open(fn, FXStreamLoad)) {
      while(stream.status() == FXStreamOK) {
        parent = tree->fromData(&stream, indent, parent, NULL, NULL, NULL);
        indent += 2;
      }
      stream.close();
    }
  }
}


long LAEWin::onMnuSave(FXObject* sender, FXSelector sel, void* ptr)
{
  FXString fn;
  
  if(!save()) return 0;
  return 1;
}


FXbool LAEWin::save()
{
  FXFileStream stream;
  FXTreeItem* item = 0;
  
  if(fileName.empty()) return saveAs();
  
  tree->setFocus();
  if(!stream.open(fileName, FXStreamSave)) return FALSE;
  
  for(item = tree->getFirstItem(); item; item = item->getNext()) {
    tree->toData(&stream, item, "");
  }
  stream.close();
  
  return TRUE;
}


long LAEWin::onMnuSaveAs(FXObject* sender, FXSelector sel, void* ptr)
{
  if(!saveAs()) return 0;
  return 1;
}


FXbool LAEWin::saveAs()
{
  FXString fn;
  
  fn = execDlg("Save File", TRUE);
  if(fn.empty()) return FALSE;
  
  if(fn.findf('.') < 0) fn += ".lae";
  fileName = fn;
  save();
  return TRUE;
}


long LAEWin::onMnuAbout(FXObject* sender, FXSelector sel, void* ptr)
{
  FXString msg;
  
  msg.format(
    APPNAME" ver."VERSION"\n"
    APPNAME" is in the public domain.\n"
    "Developed by "AUTHOR"\n"
    "\n"
    "Using FOX Library Version %d.%d.%d\n"
    "Copyright (C) 2000,2001 Jeroen van der Zijp <jeroen@fox-toolkit.org>\n"
    "http://www.fox-toolkit.org/",
    FOX_MAJOR, FOX_MINOR, FOX_LEVEL
  );
  FXMessageBox about(this, "About "APPNAME, msg, NULL, MBOX_OK|DECOR_TITLE|DECOR_BORDER);
  about.execute();
  tree->setFocus();
  return 1;
}


long LAEWin::onSetCurrent(FXObject* sender, FXSelector sel, void* ptr)
{
  FXTreeItem* c = 0;
  FXString str;
  
  c = tree->getFirstItem();
  switch((FXint)ptr) {
  case 1:
    for(; c; c = c->getNext()) str += genC(c);
    codeC->setText(str);
    break;
  case 2:
    for(; c; c = c->getNext()) str += genPerl(c);
    codePerl->setText(str);
    break;
  case 3:
    for(; c; c = c->getNext()) str += genPython(c);
    codePython->setText(str);
    break;
  case 4:
    for(; c; c = c->getNext()) str += genRuby(c);
    codeRuby->setText(str);
    break;
  case 5:
    for(; c; c = c->getNext()) str += genVB(c);
    codeVB->setText(str);
    break;
  case 6:
    for(; c; c = c->getNext()) str += genPascal(c);
    codePascal->setText(str);
    break;
  case 7:
    for(; c; c = c->getNext()) str += genTcl(c);
    codeTcl->setText(str);
    break;
  }
  return 1;
}


FXString LAEWin::genC(FXTreeItem* item, const FXString& indent /*= ""*/, const FXString& prefix /*= ""*/)
{
  FXString text;
  FXTreeItem* p = 0;
  FXTreeItem* c = 0;
  FXString pfx;
  FXString ret;
  
  if(!item) return "";
  
  p = item->getParent();
  c = item->getFirst();
  text = item->getText();
  if(prefix == "if") {
    ret = indent + "if(" + text + ")";
    if(c) {
      ret += " {\n";
      for(; c; c = c->getNext()) {
        ret += genC(c, indent + "  ");
      }
    } else {
      ret += ";\n";
    }
  } else if(prefix == "else") {
    ret = indent + "} else";
    if(!text.empty()) ret += " if(" + text + ")";
    if(c) {
      ret += " {\n";
      for(; c; c = c->getNext()) {
        ret += genC(c, indent + "  ");
      }
    } else {
      ret += ";\n";
    }
  } else if(text == "if") {
    if(c) {
      pfx = "if";
      for(; c; c = c->getNext()) {
        ret += genC(c, indent, pfx);
        pfx = "else";
      }
      ret += indent + "}\n";
    } else {
      ret = "/* if */\n";
    }
  } else {
    if(!text.empty()) {
      ret = indent + text + ";\n";
    } else {
      ret = indent + "\n";
    }
    for(; c; c = c->getNext()) {
      ret += genC(c, indent);
    }
  }
  return ret;
}


FXString LAEWin::genPerl(FXTreeItem* item, const FXString& indent /*= ""*/, const FXString& prefix /*= ""*/)
{
  FXString text;
  FXTreeItem* p = 0;
  FXTreeItem* c = 0;
  FXString pfx;
  FXString ret;
  
  if(!item) return "";
  
  p = item->getParent();
  c = item->getFirst();
  text = transPerl(item->getText());
  if(prefix == "if") {
    ret = indent + "if(" + text + ")";
    if(c) {
      ret += " {\n";
      for(; c; c = c->getNext()) {
        ret += genPerl(c, indent + "  ");
      }
    } else {
      ret += ";\n";
    }
  } else if(prefix == "else") {
    if(text.empty()) {
      ret = indent + "} else";
    } else {
      ret = indent + "} elsif(" + text + ")";
    }
    if(c) {
      ret += " {\n";
      for(; c; c = c->getNext()) {
        ret += genPerl(c, indent + "  ");
      }
    } else {
      ret += ";\n";
    }
  } else if(text == "if") {
    if(c) {
      pfx = "if";
      for(; c; c = c->getNext()) {
        ret += genPerl(c, indent, pfx);
        pfx = "else";
      }
      ret += indent + "}\n";
    } else {
      ret = "# if\n";
    }
  } else {
    if(!text.empty()) {
      ret = indent + text + ";\n";
    } else {
      ret = indent + "\n";
    }
    for(; c; c = c->getNext()) {
      ret += genPerl(c, indent);
    }
  }
  return ret;
}


FXString LAEWin::transPerl(const FXString& text)
{
  int i = 0;
  FXString sign;
  int len = 0;
  char ch = 0;
  FFCString str;
  int type = 0;
  int oldt = 0;
  FFCList list;
  int sz = 0;
  FFCString ret;
  FFCString** slist = 0;
  
  list.setDeleteFunc(FFCList::delString);
  list.setCopyFunc(FFCList::copyString);
  
  len = text.length();
  for(i = 0; i <= len; i++) {
    ch = text[i];
    oldt = type;
    if(ch == ' ' || ch == '\0') {
      type = 0;
    } else if(('A' <= ch && ch <= 'Z') || ('a' <= ch && ch <= 'z') || ch == '_') {
      type = 1;
    } else if(('0' <= ch && ch <= '9') || ch == '.') {
      if(type != 1) type = 2;
    } else {
      type = 3;
    }
    if(type != oldt && !str.empty()) {
      if(oldt == 1 && str != "if") str = "$" + str;
      list += &str;
      str.erase();
    }
    if(type != 0) str += ch;
  }
  
  sz = list.length();
  slist = (FFCString**)list.data();
  for(i = 0; i < sz; i++) {
    if(i > 0 && *slist[i - 1] != "!" && *slist[i] != "++" && *slist[i] != "--") ret += ' ';
    ret += *slist[i];
  }
  return ret.data();
}


FXString LAEWin::genPython(FXTreeItem* item, const FXString& indent /*= ""*/, const FXString& prefix /*= ""*/)
{
  FXString text;
  FXTreeItem* p = 0;
  FXTreeItem* c = 0;
  FXString pfx;
  FXString ret;
  
  if(!item) return "";
  
  p = item->getParent();
  c = item->getFirst();
  text = transPython(item->getText());
  if(prefix == "if") {
    ret = indent + "if " + text + ":\n";
    for(; c; c = c->getNext()) {
      ret += genPython(c, indent + "  ");
    }
  } else if(prefix == "else") {
    if(text.empty()) {
      ret = indent + "else:\n";
    } else {
      ret = indent + "elif " + text + ":\n";
    }
    for(; c; c = c->getNext()) {
      ret += genPython(c, indent + "  ");
    }
  } else if(text == "if") {
    if(c) {
      pfx = "if";
      for(; c; c = c->getNext()) {
        ret += genPython(c, indent, pfx);
        pfx = "else";
      }
    } else {
      ret = "# if\n";
    }
  } else {
    ret = indent + text + "\n";
    for(; c; c = c->getNext()) {
      ret += genPython(c, indent);
    }
  }
  return ret;
}


FXString LAEWin::transPython(const FXString& text)
{
  int i = 0;
  FXString sign;
  int len = 0;
  char ch = 0;
  FFCString str;
  int type = 0;
  int oldt = 0;
  FFCList list;
  int sz = 0;
  FFCString ret;
  FFCString** slist = 0;
  
  list.setDeleteFunc(FFCList::delString);
  list.setCopyFunc(FFCList::copyString);
  
  len = text.length();
  for(i = 0; i <= len; i++) {
    ch = text[i];
    oldt = type;
    if(ch == ' ' || ch == '\0') {
      type = 0;
    } else if(('A' <= ch && ch <= 'Z') || ('a' <= ch && ch <= 'z') || ch == '_') {
      type = 1;
    } else if(('0' <= ch && ch <= '9') || ch == '.') {
      if(type != 1) type = 2;
    } else {
      type = 3;
    }
    if(type != oldt && !str.empty()) {
      list += &str;
      str.erase();
    }
    if(type != 0) str += ch;
  }
  
  sz = list.length();
  slist = (FFCString**)list.data();
  for(i = 0; i < sz; i++) {
    if(i > 0 && *slist[i - 1] != "!") ret += ' ';
    if(*slist[i] == "++") {
      ret += "+= 1";
    } else if(*slist[i] == "--") {
      ret += "-= 1";
    } else if(*slist[i] == "&&") {
      ret += "and";
    } else if(*slist[i] == "||") {
      ret += "or";
    } else {
      ret += *slist[i];
    }
  }
  return ret.data();
}


FXString LAEWin::genRuby(FXTreeItem* item, const FXString& indent /*= ""*/, const FXString& prefix /*= ""*/)
{
  FXString text;
  FXTreeItem* p = 0;
  FXTreeItem* c = 0;
  FXString pfx;
  FXString ret;
  
  if(!item) return "";
  
  p = item->getParent();
  c = item->getFirst();
  text = transRuby(item->getText());
  if(prefix == "if") {
    ret = indent + "if " + text + "\n";
    for(; c; c = c->getNext()) {
      ret += genRuby(c, indent + "  ");
    }
  } else if(prefix == "else") {
    if(text.empty()) {
      ret = indent + "else\n";
    } else {
      ret = indent + "elsif " + text + "\n";
    }
    for(; c; c = c->getNext()) {
      ret += genRuby(c, indent + "  ");
    }
  } else if(text == "if") {
    if(c) {
      pfx = "if";
      for(; c; c = c->getNext()) {
        ret += genRuby(c, indent, pfx);
        pfx = "else";
      }
      ret += indent + "end\n";
    } else {
      ret = "# if\n";
    }
  } else {
    ret = indent + text + "\n";
    for(; c; c = c->getNext()) {
      ret += genRuby(c, indent);
    }
  }
  return ret;
}


FXString LAEWin::transRuby(const FXString& text)
{
  int i = 0;
  FXString sign;
  int len = 0;
  char ch = 0;
  FFCString str;
  int type = 0;
  int oldt = 0;
  FFCList list;
  int sz = 0;
  FFCString ret;
  FFCString** slist = 0;
  
  list.setDeleteFunc(FFCList::delString);
  list.setCopyFunc(FFCList::copyString);
  
  len = text.length();
  for(i = 0; i <= len; i++) {
    ch = text[i];
    oldt = type;
    if(ch == ' ' || ch == '\0') {
      type = 0;
    } else if(('A' <= ch && ch <= 'Z') || ('a' <= ch && ch <= 'z') || ch == '_') {
      type = 1;
    } else if(('0' <= ch && ch <= '9') || ch == '.') {
      if(type != 1) type = 2;
    } else {
      type = 3;
    }
    if(type != oldt && !str.empty()) {
      list += &str;
      str.erase();
    }
    if(type != 0) str += ch;
  }
  
  sz = list.length();
  slist = (FFCString**)list.data();
  for(i = 0; i < sz; i++) {
    if(i > 0 && *slist[i - 1] != "!") ret += ' ';
    if(*slist[i] == "++") {
      ret += "+= 1";
    } else if(*slist[i] == "--") {
      ret += "-= 1";
    } else {
      ret += *slist[i];
    }
  }
  return ret.data();
}


FXString LAEWin::genVB(FXTreeItem* item, const FXString& indent /*= ""*/, const FXString& prefix /*= ""*/)
{
  FXString text;
  FXTreeItem* p = 0;
  FXTreeItem* c = 0;
  FXString pfx;
  FXString ret;
  
  if(!item) return "";
  
  p = item->getParent();
  c = item->getFirst();
  text = transVB(item->getText());
  if(prefix == "if") {
    ret = indent + "If " + text + " Then\n";
    for(; c; c = c->getNext()) {
      ret += genVB(c, indent + "  ");
    }
  } else if(prefix == "else") {
    if(text.empty()) {
      ret = indent + "Else\n";
    } else {
      ret = indent + "ElseIf " + text + " Then\n";
    }
    for(; c; c = c->getNext()) {
      ret += genVB(c, indent + "  ");
    }
  } else if(text == "if") {
    if(c) {
      pfx = "if";
      for(; c; c = c->getNext()) {
        ret += genVB(c, indent, pfx);
        pfx = "else";
      }
      ret += indent + "End If\n";
    } else {
      ret = "' If\n";
    }
  } else {
    ret = indent + text + "\n";
    for(; c; c = c->getNext()) {
      ret += genVB(c, indent);
    }
  }
  return ret;
}


FXString LAEWin::transVB(const FXString& text)
{
  int i = 0;
  FXString sign;
  int len = 0;
  char ch = 0;
  FFCString str;
  int type = 0;
  int oldt = 0;
  FFCList list;
  int sz = 0;
  FFCString ret;
  FFCString** slist = 0;
  
  list.setDeleteFunc(FFCList::delString);
  list.setCopyFunc(FFCList::copyString);
  
  len = text.length();
  for(i = 0; i <= len; i++) {
    ch = text[i];
    oldt = type;
    if(ch == ' ' || ch == '\0') {
      type = 0;
    } else if(('A' <= ch && ch <= 'Z') || ('a' <= ch && ch <= 'z') || ch == '_') {
      type = 1;
    } else if(('0' <= ch && ch <= '9') || ch == '.') {
      if(type != 1) type = 2;
    } else {
      type = 3;
    }
    if(type != oldt && !str.empty()) {
      list += &str;
      str.erase();
    }
    if(type != 0) str += ch;
  }
  
  sz = list.length();
  slist = (FFCString**)list.data();
  for(i = 0; i < sz; i++) {
    if(i > 0 && *slist[i - 1] != "!") ret += ' ';
    if(i > 0 && *slist[i] == "++") {
      ret += "= " + *slist[i - 1] + " + 1";
    } else if(i > 0 && *slist[i] == "--") {
      ret += "= " + *slist[i - 1] + " - 1";
    } else if(*slist[i] == "==") {
      ret += "=";
    } else if(*slist[i] == "!=") {
      ret += "<>";
    } else if(*slist[i] == "&&") {
      ret += "And";
    } else if(*slist[i] == "||") {
      ret += "Or";
    } else {
      ret += *slist[i];
    }
  }
  return ret.data();
}


FXString LAEWin::genPascal(FXTreeItem* item, const FXString& indent /*= ""*/, const FXString& prefix /*= ""*/)
{
  FXString text;
  FXTreeItem* p = 0;
  FXTreeItem* c = 0;
  FXString pfx;
  FXString ret;
  
  if(!item) return "";
  
  p = item->getParent();
  c = item->getFirst();
  text = transPascal(item->getText());
  if(prefix == "if") {
    ret = indent + "if " + text + " then begin\n";
    for(; c; c = c->getNext()) {
      ret += genPascal(c, indent + "  ");
    }
  } else if(prefix == "else") {
    ret = indent + "end else";
    if(!text.empty()) ret += " if " + text + " then";
    ret += " begin\n";
    for(; c; c = c->getNext()) {
      ret += genPascal(c, indent + "  ");
    }
  } else if(text == "if") {
    if(c) {
      pfx = "if";
      for(; c; c = c->getNext()) {
        ret += genPascal(c, indent, pfx);
        pfx = "else";
      }
      ret += indent + "end;\n";
    } else {
      ret = "{ if }\n";
    }
  } else {
    if(!text.empty()) {
      ret = indent + text + ";\n";
    } else {
      ret = indent + "\n";
    }
    for(; c; c = c->getNext()) {
      ret += genPascal(c, indent);
    }
  }
  return ret;
}


FXString LAEWin::transPascal(const FXString& text)
{
  int i = 0;
  FXString sign;
  int len = 0;
  char ch = 0;
  FFCString str;
  int type = 0;
  int oldt = 0;
  FFCList list;
  int sz = 0;
  FFCString ret;
  FFCString** slist = 0;
  
  list.setDeleteFunc(FFCList::delString);
  list.setCopyFunc(FFCList::copyString);
  
  len = text.length();
  for(i = 0; i <= len; i++) {
    ch = text[i];
    oldt = type;
    if(ch == ' ' || ch == '\0') {
      type = 0;
    } else if(('A' <= ch && ch <= 'Z') || ('a' <= ch && ch <= 'z') || ch == '_') {
      type = 1;
    } else if(('0' <= ch && ch <= '9') || ch == '.') {
      if(type != 1) type = 2;
    } else {
      type = 3;
    }
    if(type != oldt && !str.empty()) {
      list += &str;
      str.erase();
    }
    if(type != 0) str += ch;
  }
  
  sz = list.length();
  slist = (FFCString**)list.data();
  for(i = 0; i < sz; i++) {
    if(i > 0 && *slist[i - 1] != "!") ret += ' ';
    if(i > 0 && *slist[i] == "++") {
      ret += ":= " + *slist[i - 1] + " + 1";
    } else if(i > 0 && *slist[i] == "--") {
      ret += ":= " + *slist[i - 1] + " - 1";
    } else if(*slist[i] == "=") {
      ret += ":=";
    } else if(*slist[i] == "==") {
      ret += "=";
    } else if(*slist[i] == "!=") {
      ret += "<>";
    } else if(*slist[i] == "&&") {
      ret += "and";
    } else if(*slist[i] == "||") {
      ret += "or";
    } else {
      ret += *slist[i];
    }
  }
  return ret.data();
}


FXString LAEWin::genTcl(FXTreeItem* item, const FXString& indent /*= ""*/, const FXString& prefix /*= ""*/)
{
  FXString text;
  FXTreeItem* p = 0;
  FXTreeItem* c = 0;
  FXString pfx;
  FXString ret;
  
  if(!item) return "";
  
  p = item->getParent();
  c = item->getFirst();
  text = transTcl(item->getText());
  if(prefix == "if") {
    ret = indent + "if { " + text + " } ";
    ret += " {\n";
    for(; c; c = c->getNext()) {
      ret += genTcl(c, indent + "  ");
    }
  } else if(prefix == "else") {
    if(text.empty()) {
      ret = indent + "} else";
    } else {
      ret = indent + "} elseif { " + text + " } ";
    }
    ret += " {\n";
    for(; c; c = c->getNext()) {
      ret += genTcl(c, indent + "  ");
    }
  } else if(text == "if") {
    if(c) {
      pfx = "if";
      for(; c; c = c->getNext()) {
        ret += genTcl(c, indent, pfx);
        pfx = "else";
      }
      ret += indent + "}\n";
    } else {
      ret = "# if\n";
    }
  } else {
    ret = indent + text + "\n";
    for(; c; c = c->getNext()) {
      ret += genTcl(c, indent);
    }
  }
  return ret;
}


FXString LAEWin::transTcl(const FXString& text)
{
  int i = 0;
  FXString sign;
  int len = 0;
  char ch = 0;
  FFCString str;
  int type = 0;
  int oldt = 0;
  FFCList list;
  int sz = 0;
  FFCString ret;
  FFCString** slist = 0;
  FFCString var;
  
  list.setDeleteFunc(FFCList::delString);
  list.setCopyFunc(FFCList::copyString);
  
  len = text.length();
  for(i = 0; i <= len; i++) {
    ch = text[i];
    oldt = type;
    if(ch == ' ' || ch == '\0') {
      type = 0;
    } else if(('A' <= ch && ch <= 'Z') || ('a' <= ch && ch <= 'z') || ch == '_') {
      type = 1;
    } else if(('0' <= ch && ch <= '9') || ch == '.') {
      if(type != 1) type = 2;
    } else {
      type = 3;
    }
    if(type != oldt && !str.empty()) {
      if(oldt == 1 && str != "if") str = "$" + str;
      list += &str;
      str.erase();
    }
    if(type != 0) str += ch;
  }
  
  sz = list.length();
  slist = (FFCString**)list.data();
  for(i = 0; i < sz; i++) {
    if(i > 0 && *slist[i - 1] != "!" && *slist[i] != "++" && *slist[i] != "--") ret += ' ';
    if(slist[i]->left(1) == "$" && i < sz - 1) {
      var = slist[i]->mid(1, slist[i]->length() - 1);
      if(*slist[i + 1] == "=") {
        ret += "set " + var;
        i++;
        continue;
      } else if(*slist[i + 1] == "++") {
        ret += "set " + var + " [expr $" + var + " + 1]";
        i++;
        continue;
      } else if(*slist[i + 1] == "--") {
        ret += "set " + var + " [expr $" + var + " - 1]";
        i++;
        continue;
      }
    }
    ret += *slist[i];
  }
  return ret.data();
}
